/*{{{C}}}*//* #includes */ /*{{{*/
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/kd.h>
#include <sys/stat.h>
#include <sys/vt.h>
#include <sys/sysmacros.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
/*}}}*/
/* #defines */ /*{{{*/
/* video ports */
#define INDEX_REG  0x3b4
#define DATA_REG   0x3b5
#define MODE_REG   0x3b8
#define CONFIG_REG 0x3bf

/* configurations */
#define TEXT_CONF 1
#define GRAPH_CONF 3

/* modes */
#define ACTIVE_MODE 0x08
#define TEXT_MODE 0x20
#define GRAPH_MODE0 0x02
#define GRAPH_MODE1 0x82

/* memory addresses */
#define GRAPH_MODE0_BASE 0xB0000
#define GRAPH_MODE1_BASE 0xB8000
#define GRAPH_SIZE 0x8000
/*}}}*/

/* variables */ /*{{{*/
char *pout;
static int console_fd;
static int console_nr;
/*}}}*/
/* port access */ /*{{{*/
inline static void port_out(char value, unsigned short port)
{
__asm__ volatile ("outb %0,%1"
		::"a" ((char) value),"d" ((unsigned short) port));
}


inline static unsigned char port_in(unsigned short port)
{
	unsigned char _v;
__asm__ volatile ("inb %1,%0"
		:"=a" (_v):"d" ((unsigned short) port));
	return _v;
}
/*}}}*/

/* mapscreen */ /*{{{*/
unsigned char *mapscreen(void)
{
  /* variables */ /*{{{*/
  struct stat buf;
  unsigned char *vmem;
  int mem_fd;
  /*}}}*/

  if (geteuid()>0) setreuid(geteuid(),getuid());
  /* open console */ /*{{{*/
  if ((console_fd=open("/dev/console",O_RDWR))==-1)
  {
    setreuid(0,getuid());
    pout="can't open console";
    return((unsigned char*)0);
  }
  if (stat("/dev/tty",&buf)==-1)
  {
    setreuid(0,getuid());
    pout="can't stat tty";
    return((unsigned char*)0);
  }
  console_nr=minor(buf.st_rdev);
  /*}}}*/
  /* open /dev/mem */ /*{{{*/
  if ((mem_fd=open("/dev/mem",O_RDWR))<0)
  {
    pout="can't open /dev/mem";
    setreuid(0,getuid());
    return((unsigned char*)0);
  }
  /*}}}*/
  /* mmap screen */ /*{{{*/
  if ((vmem=malloc(GRAPH_SIZE+(getpagesize())))==(unsigned char*)0)
  {
    pout="malloc() failed";
    setreuid(0,getuid());
    return((unsigned char*)0);
  }
  if ((unsigned long)vmem%getpagesize()) vmem+=getpagesize()-((unsigned long)vmem%getpagesize());
  vmem=(unsigned char*)mmap((caddr_t)vmem,GRAPH_SIZE,PROT_READ|PROT_WRITE,MAP_SHARED|MAP_FIXED,mem_fd,GRAPH_MODE1_BASE);
  if ((long)vmem<0)
  {
    pout="mmap() failed";
    setreuid(0,getuid());
    return((unsigned char*)0);
  }
  /*}}}*/
  /* get I/O permissions for HGC registers */ /*{{{*/
  if (ioperm(INDEX_REG, 1, 1))
  {
    pout="can't set register";
    setreuid(0,getuid());
    return ((unsigned char*)0);
  }
  ioperm(DATA_REG, 1, 1);
  ioperm(MODE_REG,  1, 1);
  ioperm(CONFIG_REG,  1, 1);
  /*}}}*/
  setreuid(0,getuid());
  return vmem;
}
/*}}}*/
/* grafmode */ /*{{{*/
void grafmode(void)
{
  /* variables */ /*{{{*/
  static char grfval[12]={0x35,0x2d,0x2e,0x7,0x5b,0x02,0x57,0x57,0x2,0x3,0x0,0x0};
  int regsel;
  /*}}}*/

  if (geteuid()>0) setreuid(geteuid(),getuid());
  /* block vc -- more a hack ... */ /*{{{*/
  ioctl(console_fd,VT_ACTIVATE,console_nr);
  ioctl(console_fd,VT_WAITACTIVE,console_nr);
  ioctl(console_fd,KDSETMODE,KD_GRAPHICS);
  /*}}}*/
  /* program 6845 for graphics mode */ /*{{{*/
  port_out(GRAPH_MODE1,MODE_REG);
  for (regsel=0; regsel<12; regsel++)
  {
    port_out(regsel,INDEX_REG);
    port_out(grfval[regsel],DATA_REG);
  }
  port_out(GRAPH_CONF,CONFIG_REG);
  port_out(GRAPH_MODE1|ACTIVE_MODE,MODE_REG);
  /*}}}*/
  setreuid(0,getuid());
}
/*}}}*/
/* textmode */ /*{{{*/
void textmode(void)
{
  /* variables */ /*{{{*/
  static char txtval[12]={0x61,0x50,0x52,0xf,0x19,0x06,0x19,0x19,0x2,0xd,0xb,0xc};
  int regsel;
  /*}}}*/

  if (geteuid()>0) setreuid(geteuid(),getuid());
  /* switch 6845 to text mode */ /*{{{*/
  port_out(TEXT_MODE,MODE_REG);
  for (regsel=0; regsel<12; regsel++)
  {
    port_out(regsel,INDEX_REG);
    port_out(txtval[regsel],DATA_REG);
  }
  port_out(TEXT_CONF,CONFIG_REG);
  port_out(TEXT_MODE|ACTIVE_MODE,MODE_REG);
  /*}}}*/
  /* release vt */ /*{{{*/
  ioctl(console_fd,KDSETMODE,KD_TEXT);
  /*}}}*/
  setreuid(0,getuid());
}
/*}}}*/
